from typing import Optional
import logging
import numpy as np

from scipy import optimize as opt

logger = logging.getLogger(__name__)

class Optimizer:
    def __init__(self, *args):
        pass

    def optimize(self, objective_function):
        pass